&& HYPER - A simple Hypertext system for TDBS
&& ==========================================
&& HYPER.PRG - the main program
&&
&& Released to the public domain by Dan Lyke (danlyke@flutterby.com)









PUBLIC gsFileRoot
PUBLIC gsFName
PUBLIC gsTermType
PUBLIC gfGetOkay
PUBLIC ghBackground



gfGetOkay = .F.

gsTermType = "ANSI"

ghBackground = -1

gsCNoLink = "(no link selected)"
gsCLink = "Link:"
gsCSearchPrompt = "Enter the text to search for"
gsCSearching = "Searching..."
gsCRepeating = "Repeating last search..."
gsCFind = "Enter the title to find"

DO LoadKeys



&& If the optional data exists, then we want to parse out the filename
&& from it, otherwise we go for the filename straight out. Note that
&& your subdirectory names CAN NOT have any "."s in them as the code to
&& get the root is rather simplistic. This can be enhanced...

IF CHR(38)+CHR(38) $ OPTDATA()
	gsFName = LTRIM(RTRIM(SUBSTR(OPTDATA(), AT(CHR(38)+CHR(38), OPTDATA()) + 2)))

	DO GetOptions

	IF "." $ gsFName
		gsFileRoot = LEFT(gsFName, AT(".", gsFName) - 1)
	ELSE
		gsFileRoot = gsFName
	ENDIF
ELSE
	gsFileRoot = "HYPER"
	gsFName = gsFileRoot + ".HTM"
ENDIF


IF .NOT. UANSI()
	IF gsTermType = "RIP"
	ELSE
		DO FatalError WITH "You must be running RIP or ANSI for HYPER! to work."
	ENDIF
ENDIF



&& Now that we've set the terminal type, we can load 
&& the metacharacters and attribute strings (which 
&& change dependent upon the metacharacter data)

DO LoadMetachars
DO LoadAttributes WITH gsTermType




DO CASE
	CASE gsTermType = "ANSI"
		FOPEN ghBackground "HYPERBKG.ANS" 10
	CASE gsTermType = "RIP"
		FOPEN ghBackground "HYPERBKG.RIP" 10
		DO OpenRIP
	OTHERWISE
ENDCASE

SET ESCAPE OFF
DO Splash

DO CASE
	CASE gsTermType = "ANSI"
		CLEAR
	CASE gsTermType = "RIP"
		DO CloseRIP
		CLEAR
	OTHERWISE
ENDCASE


DO CondRebuild

DO CASE
	CASE gsTermType = "ANSI"
		CLEAR
		DO Browse WITH UMORE() - 3
		CLEAR
	CASE gsTermType = "RIP"
		CLEAR
		DO OpenRIP
		DO BeginRIP
		DO Browse WITH 26
		DO EndRIP
		DO CloseRIP
	OTHERWISE
		DO Browse WITH UMORE() - 3
ENDCASE









&& PROCEDURE GetOptions
&& ======================================
&&
&& Get and process the options off of the optional data command
&& line. Done here so that some of our memory is dealt with much
&& better.

PROCEDURE GetOptions
	PRIVATE nWhere, sOption

	nWhere = AT(" ", gsFName)

	DO WHILE nWhere > 0
		sOption = RTRIM(LEFT(gsFName, nWhere))
		gsFName = LTRIM(SUBSTR(gsFName, nWhere))

		DO CASE 
			CASE sOption = "TERM=RIP"
				gsTermType = "RIP"
			CASE sOption = "TERM=ANSI"
				gsTermType = "ANSI"
			CASE sOption = "GET=YES"
				gfGetOkay = .T.
			CASE sOption = "GET=NO"
				gfGetOkay = .F.
		ENDCASE
		nWhere = AT(" ", gsFName)
	ENDDO
RETURN



&& PROCEDURE FatalError
&& PARAMETERS sString
&& ======================================
&& Print out the message as a fatal error an exit the program. At some
&& point it might be worthwhile to log a message so that the sysop can
&& tell what went wrong and why.

PROCEDURE FatalError
PARAMETERS sString

	DO CASE
		CASE gsTermType = "ANSI"
			CLEAR
		CASE gsTermType = "RIP"
			? "!|*"
			? ""
		OTHERWISE
	ENDCASE

	? "Fatal Error:"
	?? sString
	WAIT
	QUIT
RETURN




&& PROCEDURE BAddToOutput
&& PARAMETERS sString
&& ======================================
&& Add the string to the output, updating the length appropriately and
&& substituting in the metacharacters where appropriate.
&&
&& As far as I can tell this is a major bottleneck. I'm looking at ways
&& to improve the metacharacter commands, including possibly using
&& APOS() and two arrays to give the metacommand and the replacement
&& character. I am, however, concerned about the 6k memory limit.

PROCEDURE BAddToOutput
PARAMETERS sString
PRIVATE nWhere, nTemp

	nWhere = AT("&", sString)

	DO WHILE nWhere > 0
		DO CASE
			CASE gsTermType = "ANSI"
				?? LEFT(sString, nWhere - 1)
			CASE gsTermType = "RIP"
				DO RIPText WITH LEFT(sString, nWhere - 1)
			OTHERWISE
				?? LEFT(sString, nWhere - 1)
		ENDCASE

		nLen = nLen + nWhere - 1
		sString = RIGHT(sString,LEN(sString) - nWhere)

		nWhere = AT(";", sString)

		IF nWhere > 0
			nTemp = ASCAN(asMetaString, LEFT(sString, nWhere - 1))

			IF nTemp > 0
				DO CASE
					CASE gsTermType = "ANSI"
						?? asMetaReplace[nTemp]
						nLen = nLen + LEN(asMetaReplace)
					CASE gsTermType = "RIP"
						DO RIPText WITH asMetaReplace[nTemp]
						nLen = nLen + LEN(asMetaReplace)
					OTHERWISE
						?? asMetaReplace[nTemp]
						nLen = nLen + LEN(asMetaReplace)
				ENDCASE
			ELSE
				IF LEFT(sString, 1) = "#" .AND. LEN(sString) > 1
					DO CASE
						CASE gsTermType = "ANSI"
							?? CHR(VAL(SUBSTR(sString,2)))
							nLen = nLen + 1
						CASE gsTermType = "RIP"
							DO RIPText WITH CHR(VAL(SUBSTR(sString,2)))
							nLen = nLen + 1
						OTHERWISE
							?? CHR(VAL(SUBSTR(sString,2)))
							nLen = nLen + 1
					ENDCASE
				ENDIF
			ENDIF
		ENDIF

		sString = RIGHT(sString,LEN(sString) - nWhere)
		nWhere = AT("&", sString)
	ENDDO

	sOutput = sOutput + sString
	nLen = nLen + LEN(sString)

	DO CASE
		CASE gsTermType = "ANSI"
			?? sString
		CASE gsTermType = "RIP"
			DO RIPText WITH sString
		OTHERWISE
			?? sString
	ENDCASE
RETURN




&& PROCEDURE BAddAttribute
&& PARAMETERS sCommand
&& ======================================
&&
&& Process the specified HTML command, without the surrounding
&& less than and greater than signs.

PROCEDURE BAddAttribute
PARAMETERS sCommand
	PRIVATE nWhere

	nWhere = ASCAN(asAttrString, sCommand)

	IF nWhere > 0
		?? asAttrReplace[nWhere]
	ENDIF

	IF "KBD" $ sCommand
		nLen = nLen + 2
	ENDIF

	IF LEFT(sCommand,1) = "H" .AND. gsTermType = "RIP"
		?? "|S0101"
		&& DO RIPBar WITH 0, (3 + nLines) * 10, 639, (3 + nLines) * 10 + 8
		DO RTHBar WITH nLines
	ENDIF
RETURN






&& PROCEDURE BRedraw
&& PARAMETERS <none>
&& ======================================
&&
&& Redraw the screen. At the start of this function, the HTM file
&& position must be set to the start of the screen. This goes down until
&& it hits a title block or the end of the screen.
&&
&& This function also makes sure that the arrays for the hypertext links
&& are updated.
&&
&& At the exit, this function should be positioned at the start of the
&& next page (if anybody ever implements word wrapping this'll have to
&& seek back to the middle of the last line...)
&&
&& If one were to make the HTML parser a little more intelligent (ie: not
&& so line oriented), one would have to keep track of what attributes
&& are in force at the beginning of each page and set them.
&&
&& Another change that would be nice would be nesting commands.

PROCEDURE BRedraw
	PRIVATE nLines, sLine, sOutput, sCommand, nLen, nWhere, nLinkStop
	PRIVATE sLowerCommand, nOldLinkCount


	&& Do initialization stuff. It is important that EXACT be
	&& set to ON because we use ASCAN which uses the flag to
	&& check for validity of the arrays.

	SET EXACT ON

	nOldLinkCount = nLinkCount

	nLinkCount = 1

	nLines = 1
	nSize = 1
	nLinkStop = -1
	sLine = ""

	DO CASE
		CASE gsTermType = "ANSI"
		CASE gsTermType = "RIP"
			DO RButtons
			?? "|S0100"
			&& DO RIPBar WITH 0, 30, 639, 300
			?? "|B000UHR8C"
			?? "|c0B"
		OTHERWISE
	ENDCASE

	DO WHILE nSize > 0 .AND. .NOT. "<A NAME=" $ UPPER(sLine) ;
		.AND. nLines < nScrLen

		FLREAD hFile nSize sLine
		sLine = CRTRIM(sLine)

		IF .NOT. "<A NAME=" $ UPPER(sLine)
			DO CASE
				CASE gsTermType = "ANSI"
					@ nLines, 0 SAY ""
				CASE gsTermType = "RIP"
					&& DO RIPMove WITH 0, (3 + nLines) * 10
					DO RTVMove WITH nLines
				OTHERWISE
					? ""
			ENDCASE


			sOutput = ""
			DO BAddAttribute WITH "RESET"
			
			nLen = 0

			DO WHILE LEN(sLine) > 0
				nWhere = AT("<",sLine)

				IF nWhere <> 0
					DO BAddToOutput WITH LEFT(sLine,nWhere - 1)

					sLine = RIGHT(sLine, LEN(sLine) - nWhere)

					nWhere = AT(">",sLine)

					IF nWhere > 0
						sLowerCommand = LEFT(sLine, nWhere - 1)
						sCommand = UPPER(sLowerCommand)
						sLine = RIGHT(sLine, LEN(sLine) - nWhere)

						DO CASE
							CASE LEFT(sCommand,2) = "A "
								IF nLinkCount <= LEN(anLinkLines)
									anLinkLines[nLinkCount] = nLines
									anLinkStarts[nLinkCount] = nLen
									anLinkWhat[nLinkCount] = RIGHT(sLowerCommand,LEN(sLowerCommand) - 2)
									anLinkText[nLinkCount] = ""
									anLinkLen[nLinkCount] = 0
									nLinkStop = LEN(sOutput)
									DO BAddAttribute WITH "A"
								ENDIF
							CASE sCommand = "/A"
								IF nLinkStop >= 0 .AND. anLinkText[nLinkCount] = ""
									anLinkText[nLinkCount] = RIGHT(sOutput,LEN(sOutput) - nLinkStop)
									anLinkLen[nLinkCount] = nLen - anLinkStarts[nLinkCount]

									IF gsTermType = "RIP"
										?? "|c0A" && Color Light Green
										&&	DO RIPLine WITH anLinkStarts[nLinkCount] * 8, (3 + anLinkLines[nLinkCount]) * 10 + 8, ;
										&&		(anLinkLen[nLinkCount] + anLinkStarts[nLinkCount]) * 8,(3 + anLinkLines[nLinkCount]) * 10 + 8
										DO RTHLine WITH anLinkStarts[nLinkCount], ;
											anLinkLen[nLinkCount] + anLinkStarts[nLinkCount], ;
											anLinkLines[nLinkCount]
										&& DO RIPMouse WITH nLinkCount + 50, ;
										&&	anLinkStarts[nLinkCount] * 8, (3 + anLinkLines[nLinkCount]) * 10, ;
										&&	(anLinkLen[nLinkCount] + anLinkStarts[nLinkCount]) * 8,(3 + anLinkLines[nLinkCount]) * 10 + 8, ;
										&&	.T., .F., ;
										&&	CHR(210)+anLinkWhat[nLinkCount] + "^M"
										DO RTMouse WITH nLinkCount, ;
											anLinkStarts[nLinkCount], ;
											anLinkLen[nLinkCount] + anLinkStarts[nLinkCount], ;
											anLinkLines[nLinkCount], ;
											CHR(210)+LTRIM(STR(nLinkCount)) + "^M"
									ENDIF

									nLinkCount = nLinkCount + 1
									DO BAddAttribute WITH "/A"
								ENDIF
								nLinkStop = -1

							OTHERWISE
								DO BAddAttribute WITH sCommand
						ENDCASE
					ELSE
						sLine = ""
					ENDIF
				ELSE
					DO BAddToOutput WITH sLine
					sLine = ""
				ENDIF
			ENDDO

			DO CASE
				CASE gsTermType = "ANSI"
					?? CHR(27) + "[K"
				CASE gsTermType = "RIP"
				OTHERWISE
			ENDCASE


			nLines = nLines + 1
		ENDIF
	ENDDO

	DO CASE
		CASE gsTermType = "ANSI"
			DO WHILE nLines < nScrLen
				@ nLines, 0 SAY CHR(27) + "[K"
				nLines = nLines + 1
			ENDDO
		CASE gsTermType = "RIP"
		OTHERWISE
	ENDCASE



	IF "<A NAME=" $ UPPER(sLine)
		fEnd = .T.
	ELSE
		fEnd = .F.
	ENDIF

	FSEEK hFile nCurrPos 0
	DO BStatus WITH (100 * ((nCurrPos - FilePos) / CharCount))
RETURN



&& PROCEDURE RButtons
&& ======================================
&&
&& Redraw the RIP Buttons

PROCEDURE RButtons
PRIVATE nNumber, sString

	IF ghBackground > -1
		FSEEK ghBackground nNumber 0 0
		FLREAD ghBackground nNumber sString
		? "!|1K"
		? ""

		DO WHILE nNumber > 0
			?? sString
			FLREAD ghBackground nNumber sString
		ENDDO
		? "!"

	ELSE
		
		&& Kill all previously defined mouse
		&& fields so that we don't have over-
		&& drawing.

		?? "|1K"

		&& Next/Prev

		&&	DO RIPButton WITH 320,  0, 376,  20, "+", "+Next", "+"
		&&	DO RIPButton WITH 377,  0, 433,  20, "-", "-Prev", "-"
		?? "|1U8W00AG0K1700<>+Next<>+|1UAH00C10K1900<>-Prev<>-"


		&& Find, Search & Repeat
		&&	DO RIPButton WITH 434,  0, 481,  20, "F", "Find", "F"
		&&	DO RIPButton WITH 482,  0, 546,  20, "S", "Search", "S"
		&&	DO RIPButton WITH 547,  0, 639,  20, "R", "Repeat", "R"

		?? "|1UC200DD0K1Y00<>Find<>F|1UDE00F60K2B00<>Search<>S|1UF700HR0K2A00<>Repeat<>R"


		&&	DO RipButton WITH 122, 305, 194, 325, "U", "Up page", "U"
		&&	DO RipButton WITH 195, 305, 283, 325, "D", "Down page", "D"
		?? "|1U3E8H5E912D00<>Up page<>U|1U5F8H7V911W00<>Down page<>D"

		&&	DO RipButton WITH 284, 305, 372, 325, "N", "Next link", "N"
		&&	DO RipButton WITH 373, 305, 461, 325, "P", "Prev link", "P"
		?? "|1U7W8HAC912600<>Next link<>N|1UAD8HCT912800<>Prev link<>P"

		&&	DO RipButton WITH 462, 305, 590, 325, "^M", "<Enter> follow", "^M"
		&&	DO RipButton WITH 591, 305, 639, 325, "H", "Help", "H"
		?? "|1UCU8HGE912M00<><Enter> follow<>^M|1UGF8HHR912000<>Help<>H"

		&&	DO RipButton WITH 0, 328, 48, 349, "B", "Back", "B"
		?? "|1U00941C9P1U00<>Back<>B"
	ENDIF

RETURN

&& PROCEDURE BBackground
&& PARAMETERS <none>
&& =================================
&& Draw the background bars and simple help text. Nothing special, just
&& in one place so we can find it.


PROCEDURE BBackground
PRIVATE nNumber, sString

	DO CASE
		CASE gsTermType = "ANSI"
			SET COLOR TO GR+/BG
			IF ghBackground > -1
				FSEEK ghBackground nNumber 0 0

				FLREAD ghBackground nNumber sString

				IF nNumber > 0
					@ 0, 0 SAY CHR(27) + "[K"
					@ 0, 0 SAY CRTRIM(sString)
				ENDIF

				FLREAD ghBackground nNumber sString

				IF nNumber > 0
					@ nScrLen, 0 SAY CHR(27) + "[K"
					@ nScrLen, 0 SAY CRTRIM(sString)
				ENDIF

				FLREAD ghBackground nNumber sString

				IF nNumber > 0
					@ nScrLen + 1, 0 SAY CHR(27) + "[K"
					@ nScrLen + 1, 0 SAY CRTRIM(sString)
				ENDIF
			ELSE
				@ 0, 0 SAY CHR(27) + "[K"
				@ 0, 40 SAY "+/- Next/Prev F-Find S-Search R-Repeat"
				@ nScrLen, 0 SAY "000% Read, U-Up page, D-Down page N-Next link P-Prev link <Enter> follow H-Help" + CHR(27) + "[K"
				@ nScrLen + 1, 0 SAY "B-back  " + gsCNoLink + CHR(27) + "[K"
				SET COLOR TO W/N
			ENDIF

		CASE gsTermType = "RIP"
			&& DO RIPFontStyle WITH 0, 0, 1
			?? "|Y00000100"

&&			DO RIPTextXY WITH 40 * 8, nScrTop * 14, "+/- Next/Prev F-Find S-Search R-Repeat"


			?? "|S0107"


			&&	DO RIPBar WITH 0, 302, 638, 349
			&&	
			&&	DO RIPBar WITH 0, 0, 638, 23
			&&	
			&&	?? "|c0F" && Color Light White
			&&	
			&&	DO RIPLine WITH 0, 301, 638, 301
			&&	DO RIPLine WITH 0, 301, 0, 303
			&&	
			&&	DO RIPLine WITH 0, 21, 638, 21
			&&	DO RIPLine WITH 0, 21, 0, 24
			&&	
			&&	
			&&	?? "|c08" && Color Dark Grey
			&&	DO RIPLine WITH 1, 303, 639, 303
			&&	DO RIPLine WITH 639, 301, 639, 303
			&&	
			&&	DO RIPLine WITH 1, 24, 639, 24
			&&	DO RIPLine WITH 639, 21, 639, 24

			?? "|B008EHQ9P|B0000HQ0N|c0F|L008DHQ8D|L008D008F|L000LHQ0L|L000L000O|c08|L018FHR8F|LHR8DHR8F|L010OHR0O|LHR0LHR0O"

			&& Color Light Blue
			&&	DO RIPTextXY WITH 0, 310, "000% Read"
			?? "|c09|@008M000% Read"

			?? "|c09" && Color Light Blue
			&& DO RIPTextXY WITH 54, 333, "(no link selected)"
			?? "|@1I99(no link selected)"

			&& The Buttons and such are readrawn
			&& in the redraw function to keep a
			&& handle on the mouse fields.
		OTHERWISE
	ENDCASE
RETURN




&& PROCEDURE BStatus
&& PARAMETERS nPercent
&& ===============================
&& Update the status displayed information on the screen. This also
&& clears the link selected display just so we don't have the previous
&& text on the screen.

PROCEDURE BStatus
PARAMETERS nPercent

	IF nPercent > 100
		nPercent = 100
	ENDIF
	IF nPercent < 0
		nPercent = 0
	ENDIF

	DO CASE
		CASE gsTermType = "ANSI"
			SET COLOR TO GR+/BG
			@ nScrLen, 0 SAY STR(INT(nPercent),3)
			@ nScrLen + 1, 8 SAY gsCNoLink + CHR(27) + "[K"
			SET COLOR TO W/N
		CASE gsTermType = "RIP"
			?? "|S0107"
			&&	DO RIPBar WITH 0, 305, 24, 325
			?? "|B008H0O91"
			?? "|c09" && Color Light Blue

			&& DO RIPTextXY WITH 0, 310, STR(INT(nPercent),3)
			?? "|c09|@008M" + STR(INT(nPercent),3)

			&& DO RIPBar WITH 49, 328, 639, 349
			?? "|B1D94HR9P"
			?? "|c09" && Color Light Blue
			&& DO RIPTextXY WITH 54, 333, "(no link selected)"
			?? "|@1I99(no link selected)"
		OTHERWISE
	ENDCASE
RETURN




&& PROCEDURE BName
&& PARAMETERS sName
&& ===============================
&& Update the name of the current title block at the top of the screen.

PROCEDURE BName
PARAMETERS sName
	DO CASE
		CASE gsTermType = "ANSI"
			SET COLOR TO GR+/BG
			@ 0, 0 SAY LEFT(sName + SPACE(40),40)
			SET COLOR TO W/N
		CASE gsTermType = "RIP"
			?? "|S0107|c0E"
			&& DO RIPBar WITH 0, 0, 319, 20
			?? "|B00008V0K"

			?? "|c09|=00000001" && Color Light Blue
			&& DO RIPTextXY WITH 0, 5, sName
			?? "|@0005" + sName
		OTHERWISE
	ENDCASE
RETURN





&& PROCEDURE BShowLinks
&& PARAMETERS nLink, fWhat
&& ===============================
&& Show the specified link. If fWhat is true, then it displays as the
&& highlighted link otherwise it displays as an available link. This
&& would have to be changed if the code were changed to allow embedding
&& links inside of other text.

PROCEDURE BShowLinks
PARAMETERS nLink, fWhat
	IF nLink > 0 .AND. nLinkCount > 1
		DO CASE
			CASE gsTermType = "ANSI"
				IF fWhat
					SET COLOR TO GR+/BG
					@ nScrLen + 1, 8 SAY gsCLink + " " + anLinkWhat[nLink] + CHR(27) + "[K"
					?? CHR(27) + "[33;41m"
				ELSE
					?? CHR(27) + "[31;40m"
				ENDIF

				@ anLinkLines[nLink], anLinkStarts[nLink] SAY anLinkText[nLink]

				IF fWhat
					?? CHR(27) + "[31;40m"
				ENDIF
			CASE gsTermType = "RIP"
				IF fWhat
					?? "|S0107"
					&& DO RIPBar WITH 49, 328, 639, 349
					&& Color Light Blue
					&&	DO RIPTextXY WITH 54, 333, "Link: " + anLinkWhat[nLink]
					?? "|B1D94HR9P|c09|@1I99" + gsCLink + " " + anLinkWhat[nLink]

					?? "|c0E" && Color Light Yellow
				ELSE
					?? "|c0A" && Color Light Green
				ENDIF

				&& DO RIPTextXY WITH anLinkStarts[nLink] * 8, (3 + anLinkLines[nLink]) * 10, anLinkText[nLink]
				&&	DO RIPLine WITH anLinkStarts[nLink] * 8, (3 + anLinkLines[nLink]) * 10 + 8, ;
				&&		(anLinkLen[nLink] + anLinkStarts[nLink]) * 8,(3 + anLinkLines[nLink]) * 10 + 8
				DO RTHLine WITH anLinkStarts[nLink], ;
					anLinkLen[nLink] + anLinkStarts[nLink], ;
					anLinkLines[nLink]

				IF fWhat
					?? "|c0D" && Color Light Cyan
				ENDIF

			OTHERWISE
		ENDCASE
	ENDIF
RETURN

&& PROCEDURE TBBSFunc
&& PARAMETERS nFunc, sParms
&& ======================================
&&
&& Clear off the screen, call the TBBS function specified in nFunc
&& with the optional data in sParams, then reload all of the files
&& and redraw the screen and return the user to where they were.

PROCEDURE TBBSFunc
PARAMETERS nFunc, sParms

	DO CASE
		CASE gsTermType = "ANSI"
			CLEAR
		CASE gsTermType = "RIP"
			DO EndRIP
			DO CloseRIP
			CLEAR
		OTHERWISE
	ENDCASE

	FCLOSE hFile
	USE

	DOTBBS TYPE nFunc OPTDATA sParms

	DO CASE
		CASE gsTermType = "ANSI"
		CASE gsTermType = "RIP"
			DO OpenRIP
			DO BeginRIP
		OTHERWISE
	ENDCASE


	sIndex1 = gsFileRoot + ".ND1"
	sIndex2 = gsFileRoot + ".ND2"

	USE &gsFileRoot INDEX &sIndex1 &sIndex2

	GOTO TOP

	gsFName = gsFileRoot + ".HTM"
	FOPEN hFile &gsFName 10

	IF hFile = -1
		DO FatalError WITH "Unable to reopen file '" + gsFName + "'"
	ENDIF
RETURN


PROCEDURE LinkTBBS
PARAMETERS sLink
	PRIVATE nFunc, sParms

	sOperation = sLink
	sOperation = LTRIM(RTRIM(RIGHT(sOperation, LEN(sOperation) - 5)))

	IF " " $ sOperation
		nFunc = VAL(LEFT(sOperation, AT(" ", sOperation)))
		sLink = LTRIM(RIGHT(sOperation, LEN(sOperation) - AT(" ", sOperation)))
	ELSE
		nFunc = VAL(sOperation)
		sLink = ""
	ENDIF

	SET COLOR TO W/B
	DO TBBSFunc WITH nFunc, sLink
	DO BBackground
	sOperation = "Redraw"
RETURN



&& PROCEDURE LinkHREF
&& PARAMETERS sLink
&& ===============================
&& Follow an HREF link. This gives more functionality than is currently
&& tested and implemented, but since this isn't a speed critical section
&& I'm not concerned.


PROCEDURE LinkHREF
PARAMETERS sLink
	PRIVATE sFileRoot, sSubdirectory

	sOperation = sLink
	sOperation = RIGHT(sOperation, LEN(sOperation) - 6)
	sOperation = LEFT(sOperation, LEN(sOperation) - 1)
	sSubdirectory = ""

	IF "\" $ gsFName
		sSubdirectory = LEFT(gsFName, RAT("\", gsFName))
	ENDIF


	IF "#" $ sOperation

		gsFName = LEFT(sOperation, AT("#", sOperation) - 1)
		sWord = SUBSTR(sOperation, AT("#", sOperation) + 1)

	ELSE
		gsFName = sOperation
		sWord = ""
	ENDIF

	sOperation = ""

	IF LEN(gsFName) > 0
		IF .NOT. (SUBSTR(gsFName,2,1) = ":")
			IF .NOT. (LEFT(gsFName, 1) = "/" .OR. LEFT(gsFName, 1) = "\")
				gsFName = sSubdirectory + gsFName
			ENDIF
		ENDIF


		IF "." $ gsFName
			sFileRoot = LEFT(gsFName, AT(".", gsFName) - 1)
		ELSE
			sFileRoot = gsFName
		ENDIF


		IF sFileRoot <> gsFileRoot .AND. FILE(gsFName)
			FCLOSE hFile
			USE

			gsFileRoot = sFileRoot

			DO CondRebuild

			sIndex1 = gsFileRoot + ".ND1"
			sIndex2 = gsFileRoot + ".ND2"

			USE &gsFileRoot INDEX &sIndex1 &sIndex2

			GOTO TOP

			FOPEN hFile &gsFName 10

			IF hFile = -1
				DO FatalError WITH "Unable to open link file '" + gsFName + "'"
			ENDIF

			FSEEK hFile nCurrPos FilePos 0
			sOperation = "Down"
		ENDIF
	ENDIF

	IF LEN(sWord) > 0
		nRec = RECNO()

		SET SOFTSEEK OFF
		SET EXACT ON
		SET ORDER TO 2

		SEEK UPPER(LEFT(sWord + SPACE(LEN(Title)),LEN(Title)))
		IF FOUND()
			FSEEK hFile nCurrPos FilePos 0

			nLinkCurr = 0
			nLinkCount = 1
			nLinkOld = 0
			nHighWater = 0
			fLooping = .T.
			fEnd = .F.
			sOperation = "Down"
		ELSE
			GOTO nRec
		ENDIF
	ENDIF
RETURN





&& PROCEDURE FollowLink
&& PARAMETERS <none>
&& =========================
&& Process a follow command on the currenlty selected link. Right now
&& it only passes HREFS up to the LinkHREF procedure.

PROCEDURE FollowLink
	IF nLinkCurr < nLinkCount .AND. nLinkCurr > 0
		DO CASE
			CASE UPPER(LEFT(anLinkWhat[nLinkCurr],5)) = "HREF="
				DO AddBackLink WITH gsFileRoot + ".HTM#" + Title
				DO LinkHREF WITH anLinkWhat[nLinkCurr]
			CASE UPPER(LEFT(anLinkWhat[nLinkCurr],5)) = "TBBS="
				DO LinkTBBS WITH anLinkWhat[nLinkCurr]
		ENDCASE
	ENDIF
RETURN



&& PROCEDURE AddBackLink
&& PARAMETERS sLink
&& =========================
&& Add a link sLink to our backwards queue. If the queue isn't full, pile
&& it on. If it is full, delete the first element from the queue and add
&& this to the end. This should be in the form of an "HREF=" statement,
&& because the "LinkHREF" statement is used to retrace our steps.


PROCEDURE AddBackLink
PARAMETERS sLink
	PRIVATE nDummy

	IF nBackHigh < LEN(asBackLink)
		nBackHigh = nBackHigh + 1
		asBackLink[nBackHigh] = sLink
	ELSE
		nDummy = ADEL(asBackLink, 1)
		asBackLink[nBackHigh] = sLink
	ENDIF
RETURN




&& PROCEDURE IsInText
&& PARAMETERS sWord
&& =========================
&& If sWord exists in the current file line (read from disk file hFile)
&& then set fLooping to ".F.". This skips all of the HTML control
&& commands within the file so as to make sure that the user is only
&& getting a search on what they can see on the screen.

PROCEDURE IsInText
PARAMETERS sWord
	PRIVATE nSize, sLine, nPos, sResult

	FSEEK hFile nCurrPos 0
	FLREAD hFile nSize sLine
	FSEEK hFile nCurrPos 0

	sLine = UPPER(CRTRIM(sLine))
	sWord = LTRIM(RTRIM(UPPER(sWord)))


	IF "<A NAME=" $ sLine
		RETURN
	ENDIF

	sResult = ""
	nPos = AT("<", sLine)

	DO WHILE nPos > 0
		sResult = sResult + LEFT(sLine, nPos - 1)
		sLine = RIGHT(sLine, LEN(sLine) - nPos)

		sLine = RIGHT(sLine, LEN(sLine) - AT(">", sLine))

		nPos = AT("<", sLine)
	ENDDO

	sResult = sResult + sLine

	fLooping = .NOT. (sWord $ sResult)
RETURN


&& PROCEDURE GetDocument
&& =========================
&&
&& Take the current document, format it as ASCII and download it 
&& to the user.


PROCEDURE GetDocument
	PRIVATE hOutFile, nOldPos, sLine, nSize, nStart, nStop, sOut, nWhich, sReplace


	FCREATE hOutFile "PRINT"+ULINE()+".TXT" 13

	IF hOutFile > -1
		DO LoadAttributes WITH ""

		SET EXACT ON
		FSEEK hFile nOldPos 0

		FSEEK hFile nCurrPos FilePos 0

		FLREAD hFile nSize sLine

		DO WHILE nSize > 0 .AND. .NOT. "<A NAME=" $ UPPER(sLine)
			&& First off, take care of the HTML
			&& directives in the source string

			nStart = AT("<", sLine)
			sOut = ""

			DO WHILE nStart > 0
				nStop = AT(">", SUBSTR(sLine, nStart))

				IF nStop > 0
					nStop = nStop + nStart - 1
					nWhich = ASCAN(asAttrString, SUBSTR(sLine,nStart + 1, nStop - nStart - 1))

					IF nWhich > 0
						sReplace = asAttrReplace[nWhich]
					ELSE
						sReplace = ""
					ENDIF


					IF nStart > 1
						IF nStop < LEN(sLine) - 1
							sOut = sOut + LEFT(sLine, nStart - 1) + sReplace
							sLine = SUBSTR(sLine, nStop + 1)
						ELSE
							sOut = sOut + LEFT(sLine, nStart - 1) + sReplace
							sLine = ""
						ENDIF
					ELSE
						IF nStop < LEN(sLine) - 1
							sOut = sOut + sReplace 
							sLine = SUBSTR(sLine, nStop + 1)
						ELSE
							sOut = sReplace
							sLine = ""
						ENDIF
					ENDIF

					nStart = AT("<", sLine)
				ELSE
					nStart = 0
				ENDIF
			ENDDO


			sLine = sOut + sLine
			sOut = ""


			&& Now take care of the metacharacters.

			nStart = AT("&", sLine)

			DO WHILE nStart > 0
				nStop = AT(";", SUBSTR(sLine, nStart))

				IF nStop > 0
					nSTop = nStop + nStart - 1
					nWhich = ASCAN(asMetaString, SUBSTR(sLine,nStart + 1, nStop - nStart - 1))

					IF nWhich > 0
						sReplace = asMetaReplace[nWhich]
					ELSE
						sReplace = ""
					ENDIF


					IF nStart > 1
						IF nStop < LEN(sLine) - 1
							sOut = sOut + LEFT(sLine, nStart - 1) + sReplace
							sLine = SUBSTR(sLine, nStop + 1)
						ELSE
							sOut = sOut + LEFT(sLine, nStart - 1) + sReplace
						ENDIF
					ELSE
						IF nStop < LEN(sLine) - 1
							sOut = sOut + sReplace
							sLine =  SUBSTR(sLine, nStop + 1)
						ELSE
							sOut = sOut + sReplace
							sLine = ""
						ENDIF
					ENDIF

					nStart = AT("&", sLine)
				ELSE
					nStart = 0
				ENDIF
			ENDDO
			sLine = sOut + sLine

			FLWRITE hOutFile nSize sLine

			FLREAD hFile nSize sLine
		ENDDO


		FLWRITE hOutFile nSize CHR(13) + CHR(10)
		FCLOSE hOutFile

		DO LoadAttributes WITH gsTermType

		PRIVATE hOutFile
		PRIVATE sLine
		PRIVATE nSize
		PRIVATE sReplace
		sLine = HOMEPATH() + "PRINT"+ULINE()+".DIR"
		IF FILE(sLine)
			ERASE sLine
		ENDIF
		FCREATE hOutFile &sLine 13
		sReplace = "PRINT" + ULINE() + ".TXT PRINT.TXT printer file" + CHR(13) + CHR(10)
		FLWRITE hOutFile nSize sReplace
		FCLOSE hOutFile
		DO TBBSFunc WITH 46, HOMEPATH() + "PRINT"+ULINE()+" /NB/NC/NL/NBC/I:"+CHR(34)+"PRINT.TXT"+CHR(34)

		FSEEK hFile nCurrPos nOldPos 0
	ENDIF
RETURN



PROCEDURE KDown
	sOperation = "Down"
	nLinkCurr = 0
RETURN

PROCEDURE KUp
	sOperation = "Up"
	nLinkCurr = 0
RETURN

PROCEDURE KHelp
	IF gsFileRoot <> "HYPERHLP" .OR. RTRIM(Title) <> gsTermType
		DO AddBackLink WITH gsFileRoot + ".HTM#" + Title
		DO LinkHREF WITH "HREF=&" + HOMEPATH() + "HYPERHLP.HTM#" + gsTermType + "&"
		DO BName WITH Title
		sOperation = "Down"
	ENDIF
RETURN

PROCEDURE KRedraw
	DO CASE
		CASE gsTermType = "ANSI"
			CLEAR
		CASE gsTermType = "RIP"
			DO EndRIP
			DO CloseRIP
			@ 0,0 SAY ""
			DO OpenRIP
			DO BeginRIP
		OTHERWISE
	ENDCASE

	DO BBackground
	sOperation = "Redraw"
	DO BName WITH Title
RETURN


PROCEDURE KPageUp
	IF .NOT. BOF()
		SET ORDER TO 0
		SKIP -1

		FSEEK hFile nCurrPos FilePos 0
		nLinkCurr = 0
		nLinkCount = 1
		nLinkOld = 0
		nHighWater = 0
		fLooping = .T.
		fEnd = .F.
		sOperation = "Down"
	ENDIF
	DO BName WITH Title
RETURN


PROCEDURE KPageDown
	SET ORDER TO 0
	SKIP 1

	IF .NOT. EOF()
		FSEEK hFile nCurrPos FilePos 0
		nLinkCurr = 0
		nLinkCount = 1
		nLinkOld = 0
		nHighWater = 0
		fLooping = .T.
		fEnd = .F.
		sOperation = "Down"
	ELSE
		SKIP -1
	ENDIF
	DO BName WITH Title
RETURN

PROCEDURE KSearch
	sWord = SPACE(40)


	DO CASE
		CASE gsTermType = "ANSI"
			@ 0, 0 GET sWord
			READ
		CASE gsTermType = "RIP"
			?? "|w00001B0M10"
			?? "|S0100"
			&& DO RIPBar WITH 0, 0, 319, 20
			?? "|B00008V0K"
			DO EndRIP
			? CHR(27) + "[0;0H" + gsCSearchPrompt
			ACCEPT "" TO sWord
		OTHERWISE
			? gsCSearchPrompt
			ACCEPT "" TO sWord
	ENDCASE

	DO CASE
		CASE gsTermType = "ANSI"
		CASE gsTermType = "RIP"
			DO BeginRIP
			?? "|w0000000000"
			?? "|S0107"
			&& DO RIPBar WITH 0, 0, 319, 20
			?? "|B00008V0K"
		OTHERWISE
	ENDCASE


	IF LASTKEY() <> 27 .AND. LEN(LTRIM(sWord)) > 0
		DO CASE
			CASE gsTermType = "ANSI"
				SET COLOR TO GR+/BG
				@ 0, 0 SAY LEFT(gsCSearching + SPACE(40),40)
			CASE gsTermType = "RIP"
				DO BName WITH gsCSearching
			OTHERWISE
		ENDCASE

		FSEEK hFile nCurrPos 0 0

		fLooping = .T.

		DO WHILE fLooping
			FLFIND hFile nCurrPos LTRIM(RTRIM(sWord)) 1

			IF nCurrPos = 0
				fLooping = .F.
			ELSE
				DO IsInText WITH sWord
			ENDIF
		ENDDO

		fLooping = .T.

		IF nCurrPos <> 0
			SET EXACT OFF
			SET SOFTSEEK ON
			SET ORDER TO 1

			SEEK nCurrPos
			SKIP -1

			FSEEK hFile nCurrPos FilePos 0

			nLinkCurr = 0
			nLinkCount = 1
			nLinkOld = 0
			nHighWater = 0
			fLooping = .T.
			fEnd = .F.
			sOperation = "Down"
		ENDIF

	ENDIF
	DO BBackground
	DO BName WITH Title

RETURN


PROCEDURE KRepeat
	IF LEN(LTRIM(RTRIM(sWord))) > 0
		DO CASE
			CASE gsTermType = "ANSI"
				SET COLOR TO GR+/BG
				@ 0, 0 SAY LEFT(gsCRepeating + SPACE(40),40)
			CASE gsTermType = "RIP"
				DO BName WITH gsCRepeating
			OTHERWISE
		ENDCASE

		FSEEK hFile nCurrPos (FilePos+CharCount-2) 0

		fLooping = .T.

		DO WHILE fLooping
			FLFIND hFile nCurrPos LTRIM(RTRIM(sWord)) 1

			IF nCurrPos = 0
				fLooping = .F.
			ELSE
				DO IsInText WITH sWord
			ENDIF
		ENDDO

		fLooping = .T.

		IF nCurrPos <> 0
			SET EXACT OFF
			SET SOFTSEEK ON
			SET ORDER TO 1

			SEEK nCurrPos
			SKIP -1

			FSEEK hFile nCurrPos FilePos 0

			nLinkCurr = 0
			nLinkCount = 1
			nLinkOld = 0
			nHighWater = 0
			fLooping = .T.
			fEnd = .F.
			sOperation = "Down"
		ENDIF

		DO BBackground
		DO BName WITH Title
	ENDIF
RETURN


PROCEDURE KFind
	SET ORDER TO 2
	SET EXACT OFF
	SET SOFTSEEK ON
	sWord = SPACE(LEN(Title))

	DO CASE
		CASE gsTermType = "ANSI"
			@ 0, 0 GET sWord
			READ
		CASE gsTermType = "RIP"
			?? "|w00001B0M10"
			?? "|S0100"
			&& DO RIPBar WITH 0, 0, 319, 20
			?? "|B00008V0K"
			DO EndRIP
			? CHR(27) + "[0;0H" + gsCFind
			ACCEPT "" TO sWord
		OTHERWISE
			? gsCFind
			ACCEPT "" TO sWord
	ENDCASE


	DO CASE
		CASE gsTermType = "ANSI"
		CASE gsTermType = "RIP"
			DO BeginRIP
			?? "|w0000000000"
		OTHERWISE
	ENDCASE

	IF LASTKEY() <> 27
		SEEK UPPER(sWord)

		IF EOF()
			SKIP -1
		ENDIF

		IF BOF()
			SKIP 1
		ENDIF

		FSEEK hFile nCurrPos FilePos 0

		nLinkCurr = 0
		nLinkCount = 1
		nLinkOld = 0
		nHighWater = 0
		fLooping = .T.
		fEnd = .F.
		sOperation = "Down"
	ENDIF
	DO BBackground
	DO BName WITH Title
RETURN



PROCEDURE KNext
	nLinkCurr = nLinkCurr + 1
RETURN

PROCEDURE KPrevious
	nLinkCurr = nLinkCurr - 1

	IF nLinkCurr <= 0
		nLinkCurr = nLinkCount - 1
	ENDIF

RETURN


PROCEDURE KBack
	IF nBackHigh > 0
		DO LinkHREF WITH "HREF=&" + asBackLink[nBackHigh] + "&"
		nBackhigh = nBackHigh - 1
		DO BName WITH Title
	ELSE
	ENDIF
RETURN

PROCEDURE KDownload
	IF gfGetOkay
		DO GetDocument
		DO BBackground
		DO BName WITH Title
		sOperation = "Redraw"
	ENDIF
RETURN

PROCEDURE KMemCheck
	DECLARE asSpace[100]

	nKey = 1

	DO WHILE nKey <= LEN(asSpace)
		? "Allocated bytes:", nKey * 80
		asSpace[nKey] = SPACE(80)
		nKey = nKey + 1
	ENDDO
RETURN


PROCEDURE KFollowLink
	DO FollowLink
	DO BName WITH Title
RETURN


PROCEDURE KQuit
	fLooping = .F.
RETURN






&& PROCEDURE Browse
&& PARAMETERS nScrLen
&& ================================
&& Do the hypertext browse thing on the screen at the specified top
&& position to the specified length + 1.
&&
&& I'm still not totally cool on how TDBS handles its symbol tables, but
&& I'm fairly sure that this has all of the variables that I can clean
&& up cleaned up by the end of the procedure.


PROCEDURE Browse
PARAMETERS nScrLen
	PRIVATE hFile, nSize, sLine, nCurrPos, nLines, ;
		nHighWater, fLooping, sOperation, nKey, fEnd, ;
		fDrawn, nRec, sWord, nFunc

	PRIVATE nLinkCount, nLinkCurr, nLinkOld
	DECLARE anPageStarts[40]
	DECLARE anLinkLines[25]
	DECLARE anLinkStarts[LEN(anLinkLines)]
	DECLARE anLinkText[LEN(anLinkLines)]
	DECLARE anLinkWhat[LEN(anLinkLines)]
	DECLARE anLinkLen[LEN(anLinkLines)]
	DECLARE asBackLink[10], nBackHigh


	IF nScrLen > 24
		nScrLen = 24
	ENDIF

	IF nScrlen < 5
		nScrLen = 20
	ENDIF

	&& Open the database and files, redraw the screen, set everything
	&& to the initial states for the loop.


	SET EXCLUSIVE OFF
	sIndex1 = gsFileRoot + ".ND1"
	sIndex2 = gsFileRoot + ".ND2"

	USE &gsFileRoot INDEX &sIndex1 &sIndex2

	GOTO TOP

	FOPEN hFile &gsFName 10

	IF hFile = -1
		DO FatalError WITH "Unable to open link file '" + gsFName + "'"
	ENDIF

	FSEEK hFile nCurrPos FilePos 0

	DO BBackground

	nLinkCurr = 0
	nLinkCount = 1
	nLinkOld = 0
	nHighWater = 0
	nBackHigh = 0
	fLooping = .T.
	fEnd = .F.
	sOperation = "Down"

	DO BName WITH Title

	sWord = SPACE(40)


	DO WHILE fLooping
		&& First, handle any requested file motion or screen
		&& redraws.
		&&
		&& "Down" puts the current location in the next
		&& element of the page start array and draws the
		&& screen.
		&&
		&& "Up" seeks to the previous number in the page
		&& start array, seeks to that location and redraws
		&& the screen.
		&&
		&& "Redraw" assumes that you've already done a
		&& "Down", seeks to the current location in the page
		&& start array and draws down from there.

		DO CASE
			CASE sOperation = "Down"
				IF nHighWater < LEN(anPageStarts) .AND. .NOT. fEnd
					nHighWater = nHighWater + 1
					FSEEK hFile nCurrPos 0
					anPageStarts[nHighWater] = nCurrPos

					DO BRedraw
					nLinkOld = -1
					nLinkCurr = 0
				ENDIF

			CASE sOperation = "Up"
				IF nHighWater > 1
					nHighWater = nHighWater - 1
					FSEEK hFile nCurrPos anPageStarts[nHighWater] 0

					DO BRedraw
					nLinkOld = -1
					nLinkCurr = 0
				ENDIF
			CASE sOperation = "Redraw"
				IF nHighWater > 0
					FSEEK hFile nCurrPos anPageStarts[nHighWater] 0
					DO BRedraw
					nLinkOld = -1
				ENDIF
		ENDCASE


		&& Handle making sure that the link values are within
		&& range and redrawing the links.

		IF nLinkCurr >= nLinkCount
			nLinkCurr = 0
		ENDIF

		IF nLinkCurr <= 0 .AND. nLinkCount > 1
			nLinkCurr = 1
		ENDIF

		IF nLinkOld <> nLinkCurr
			DO BShowLinks WITH nLinkOld, .F.
			DO BShowLinks WITH nLinkCurr, .T.
			nLinkOld = nLinkCurr
		ENDIF



		&& === GET A NEW KEY AND PROCESS THE SUCKER ===


		&& CASE nKey = ASC("D") .OR. nKey = ASC("d")
		&&  Go down one page in the current title block
		&&
		&& CASE nKey = ASC("U") .OR. nKey = ASC("u")
		&&  Go up one page in the current title block
		&&
		&& CASE nKey = 24	&& down arrow
		&&  Go down one page in the current title block
		&&
		&& CASE nKey =  5	&& up arrow
		&&  Go up one page in the current title block
		&&
		&& CASE nKey =  ASC("H") .OR. nKey =  ASC("h")
		&&  Draw the help screen, wait for a keypress and
		&&  return to where we were.
		&&
		&& CASE nKey = ASC("?")
		&&  Redraw the screen. This seems to be the TDBS
		&&  standard as I can't get "?" to work for anything
		&&  else on my menus.
		&&
		&& CASE nKey = 18 .OR. nKey = ASC("-")	&& page up
		&&  Go up to the previous record IN THE ORDER OF THE
		&&  HTM FILE. This is so that you can order things
		&&  other than alpha by title.
		&&
		&& CASE nKey = 3 .OR. nKey = ASC("+")	&& page down
		&&  Go down to the previous record IN THE ORDER OF THE
		&&  HTM FILE. This is so that you can order things
		&&  other than alpha by title.
		&&
		&& CASE nKey = ASC("S") .OR. nKey = ASC("s")
		&&  Search for a word within the text of the HTM file.
		&&  This can be dog slow and I haven't really tested
		&&  it a whole lot.
		&&
		&& CASE nKey = ASC("R") .OR. nKey = ASC("r")
		&&  Repeat the last search started by a "S" or "s".
		&&  Same caveats apply.
		&&
		&& CASE nKey = ASC("F") .OR. nKey = ASC("f")
		&&  Find by title block name, case insensitive.
		&&
		&& CASE nKey = ASC("N" ) .OR. nKey = ASC("n" )	.OR. nKey =  4 && Right arrow
		&&  Move to the next link on the screen.
		&&
		&& CASE nKey = ASC("P" ) .OR. nKey = ASC("p" )	.OR. nKey = 19 && Left arrow
		&&  Move to the previous link on the screen.
		&&
		&& CASE nKey = 13 .OR. nKey = 32
		&&  Space bar or enter follows the currently highlighted link.
		&&
		&& CASE nKey = 27
		&&  Escape cancels the wait loop.


		DO CASE
			CASE gsTermType = "ANSI"
			CASE gsTermType = "RIP"
				DO EndRIP
			OTHERWISE
		ENDCASE

		nKey = INKEY(0)

		DO CASE
			CASE gsTermType = "ANSI"
			CASE gsTermType = "RIP"
				DO BeginRIP
			OTHERWISE
		ENDCASE

		sOperation = ""


		&& CASE nKey = 210
		&&  Jump to the specified HREF

		IF nKey = 210 .AND. gsTermType = "RIP"
			sWord = ""

			nKey = INKEY(0)

			DO WHILE nKey > 31 .AND. nKey < 128
				sWord = sWord + CHR(nKey)
				nKey = INKEY(0)
			ENDDO

			IF LEN(sWord) > 0
				nLinkCurr = VAL(sWord)
				DO FollowLink
			ENDIF
			&& DO BBackground
			DO BName WITH Title
		ELSE
			IF nKey > 0 .AND. nKey < 256
				nFunc = ASCAN(gasKKeys, UPPER(CHR(nKey)))

				IF nFunc > 0 .AND. nFunc < LEN(gasKFuncs)

					IF UPPER(CHR(nKey)) = gasKKeys[nFunc]
						sWord = gasKFuncs[nFunc]
						DO &sWord
					ENDIF
				ENDIF
			ENDIF
		ENDIF

	ENDDO && end of the major user interface loop.



	FCLOSE hFile

	RELEASE anPageStarts
	RELEASE anLinkLines
	RELEASE anLinkStarts
	RELEASE anLinkText
	RELEASE anLinkWhat
	SET COLOR TO W/B
RETURN









