&& HYPER - A simple Hypertext system for TDBS
&& ==========================================
&& INDEX.PRG - the indexing routines
&&
&& Released to the public domain by Dan Lyke (danlyke@flutterby.com)




PROCEDURE FlipSlashes
PRIVATE nWhere

	nWhere = AT("/", gsFileRoot)

	DO WHILE nWhere > 0
		gsFileRoot = LEFT(gsFileRoot, nWhere - 1) + "\" + SUBSTR(gsFileRoot, nWhere + 1)
		nWhere = AT("/", gsFileRoot)
	ENDDO


	nWhere = AT("/", gsFName)

	DO WHILE nWhere > 0
		gsFName = LEFT(gsFName, nWhere - 1) + "\" + SUBSTR(gsFName, nWhere + 1)
		nWhere = AT("/", gsFName)
	ENDDO
RETURN



&& PROCEDURE CondRebuild
&& PARAMETERS <none>
&& ==================================
&&
&& Check for all of the conditions that could indicate that the indexes
&& that reference the HYPER.HTM file are out of date or non existent. If
&& we find any problems, rebuild.

&&	USE (&gsFileRoot) INDEX &(gsFileRoot + ".ND1") &(gsFileRoot + ".ND2")

PROCEDURE CondRebuild
	DO FlipSlashes

	IF .NOT. (FILE(gsFileRoot + ".DBF") .AND. FILE(gsFileRoot + ".ND1") .AND. FILE(gsFileRoot + ".ND2"))
		? "Reindexing database, please wait..."
		ERASE (gsFileRoot + ".DBF")
		ERASE (gsFileRoot + ".ND1")
		ERASE (gsFileRoot + ".ND2")
		SET EXCLUSIVE ON
		DO CreateBlank
		DO ReIndex
		SET EXCLUSIVE OFF
	ELSE

		IF FDATE(gsFileRoot + ".DBF") < FDATE(gsFileRoot + ".HTM") .OR. ;
			(FDATE(gsFileRoot + ".DBF") = FDATE(gsFileRoot + ".HTM") .AND. FTIME(gsFileRoot + ".DBF") < FTIME(gsFileRoot + ".HTM"))

			? "Reindexing database, please wait..."
			ERASE gsFileRoot + ".DBF"
			ERASE gsFileRoot + ".ND1"
			ERASE gsFileRoot + ".ND2"
			SET EXCLUSIVE ON
			DO CreateBlank
			DO ReIndex
			SET EXCLUSIVE OFF
		ENDIF
	ENDIF
RETURN



&& PROCEDURE CreateBlank
&& PARAMETERS <none>
&& ===============================
&&
&& Create a blank DBF file for indexing the HTM file.


PROCEDURE CreateBlank
	? "Creating blank database for " + gsFileRoot

	CREATE STRUCT

	USE STRUCT
	APPEND BLANK
	REPLACE FIELD_NAME WITH "Title", ;
		FIELD_TYPE WITH "C", ;
		FIELD_LEN WITH 40, ;
		FIELD_DEC WITH 0

	APPEND BLANK
	REPLACE FIELD_NAME WITH "FilePos", ;
		FIELD_TYPE WITH "N", ;
		FIELD_LEN WITH 10, ;
		FIELD_DEC WITH 0

	APPEND BLANK
	REPLACE FIELD_NAME WITH "CharCount", ;
		FIELD_TYPE WITH "N", ;
		FIELD_LEN WITH 10, ;
		FIELD_DEC WITH 0

	USE
	CREATE &gsFileRoot FROM STRUCT
	ERASE "STRUCT.DBF"
	USE &gsFileRoot
RETURN



&& PROCEDURE ReIndex
&& PARAMETERS <none>
&& ==============================
&&
&& Reindex the hypertext file. This means grabbing the file location of
&& each title block, the text from the title, and the length from each
&& of the previous files so that we can get references and searching
&& later.


PROCEDURE ReIndex
	PRIVATE hFile, nSize, sLine, nCurrPos, fStarted, nStartByte, sString
	PRIVATE sIndex1, sIndex2

	USE &gsFileRoot
	sIndex1 = gsFileRoot + ".ND1"
	sIndex2 = gsFileRoot + ".ND2"

	INDEX ON FilePos TO &sIndex1
	INDEX ON UPPER(Title) TO &sIndex2
	USE &gsFileRoot INDEX &sIndex1 &sIndex2

	FOPEN hFile (gsFileRoot + ".HTM") 10

	FLREAD hFile nSize sLine

	nStopByte = -1
	nStartByte = -1
	nLines = 0

	DO WHILE nSize > 0
		sLine = CRTRIM(sLine)

		IF "<A NAME=" $ UPPER(sLine) .AND. ;
			AT("<A NAME=",UPPER(sLine)) < AT(CHR(34) + ">",UPPER(sLine))

			FSEEK hFile nCurrPos 0

			IF nStartByte > -1
				REPLACE CharCount WITH nCurrPos - nStartByte
			ENDIF

			nStartByte = nCurrPos


			nSize = AT("<A NAME=",UPPER(sLine))
			APPEND BLANK
			REPLACE Title WITH SUBSTR(sLine,  nSize + 9,AT(CHR(34) + ">",UPPER(sLine)) - nSize - 9), ;
				FilePos WITH nCurrPos, ;
				CharCount WITH 1

			? "Adding record " + Title + " at " + STR(nCurrPos)

			nLines = 0
		ELSE
			nLines = nLines + 1
		ENDIF

		FLREAD hFile nSize sLine
	ENDDO

	IF nStartByte > -1
		FSEEK hFile nCurrPos 0 2
		REPLACE CharCount WITH nCurrPos - nStartByte
	ENDIF

	FCLOSE hFile

	USE
RETURN


